<?php
namespace Bss\GeoFix\Plugin;

use Magento\Store\Model\Store as StoreModel;
use Magento\Framework\UrlInterface;
use Magento\Store\Model\StoreSwitcherInterface;
use Magento\Store\Model\StoreManagerInterface;
use Magento\Framework\Url\EncoderInterface;
use Magento\Framework\App\RequestInterface;
use Magento\Catalog\Api\ProductRepositoryInterface;
use Magento\Catalog\Api\CategoryRepositoryInterface;
use Magento\Framework\Registry;
use Magento\UrlRewrite\Model\ResourceModel\UrlRewrite;
use Bss\GeoIPAutoSwitchStore\Helper\Data as GeoHelper;

class StoreSwitcher extends \Bss\GeoIPAutoSwitchStore\Plugin\StoreSwitcher
{
    /**
     * @var RequestInterface
     */
    protected $request;

    /**
     * @var CategoryRepositoryInterface
     */
    protected $categoryRepository;

    /**
     * @var ProductRepositoryInterface
     */
    protected $productRepository;

    /**
     * @var Registry
     */
    protected $coreRegistry;

    /**
     * @var UrlRewrite
     */
    protected $urlRewriteResource;

    /**
     * @var GeoHelper
     */
    protected $geoHelper;

    /**
     * @var UrlInterface
     */
    protected $urlBuilder;

    /**
     * StoreSwitcher constructor.
     * @param StoreModel $storeModel
     * @param UrlInterface $urlBuilder
     * @param StoreSwitcherInterface $storeSwitcher
     * @param EncoderInterface $encoder
     * @param StoreManagerInterface $storeManager
     * @param RequestInterface $request
     * @param ProductRepositoryInterface $productRepository
     * @param CategoryRepositoryInterface $categoryRepository
     * @param Registry $coreRegistry
     * @param UrlRewrite $urlRewriteResource
     * @param GeoHelper $geoHelper
     */
    public function __construct(
        StoreModel $storeModel,
        UrlInterface $urlBuilder,
        StoreSwitcherInterface $storeSwitcher,
        EncoderInterface $encoder,
        StoreManagerInterface $storeManager,
        RequestInterface $request,
        ProductRepositoryInterface $productRepository,
        CategoryRepositoryInterface $categoryRepository,
        Registry $coreRegistry,
        UrlRewrite $urlRewriteResource,
        GeoHelper $geoHelper
    ) {
        parent::__construct(
            $storeModel,
            $urlBuilder,
            $storeSwitcher,
            $encoder,
            $storeManager
        );
        $this->urlBuilder = $urlBuilder;
        $this->request = $request;
        $this->categoryRepository = $categoryRepository;
        $this->productRepository = $productRepository;
        $this->coreRegistry = $coreRegistry;
        $this->urlRewriteResource = $urlRewriteResource;
        $this->geoHelper = $geoHelper;
    }

    /**
     * @param \Magento\Store\Api\Data\StoreInterface $fromStore
     * @param \Magento\Store\Api\Data\StoreInterface $toStore
     * @param \Magento\Framework\App\RequestInterface $request
     * @return string
     * @throws \Magento\Store\Model\StoreSwitcher\CannotSwitchStoreException
     */
    public function getUrlRedirect($fromStore, $toStore, $request)
    {
        // Get redirect url
        // Redirect to origin path instead of base store view url
        // Redirect if Use Store Code in Url = Yes/No
        $currentPath = ltrim($request->getOriginalPathInfo(), '/');
        $defaultStore = $this->storeManager->getDefaultStoreView();

        // Try to resolve 404 route
        if ($this->request->getFullActionName() === 'cms_noroute_index' ||
            $this->request->getFullActionName() === '__' ||
            $this->request->getFullActionName() === 'catalog_category_view' ||
            $this->request->getFullActionName() === 'catalog_product_view') {
            $requestStoreCode = $request->getParam('___from_store');
            $switcherWs = $request->getParam('___website_switcher');
            $storeConditionId = $this->storeManager->getStore()->getId();
            if ($requestStoreCode && $switcherWs) {
                $fromStore = $this->geoHelper->getStoreByCode($requestStoreCode);
                $toStore = $this->storeManager->getStore();
                $storeConditionId = $fromStore->getId();
            }
//            $previousRoute = $request->getServer('HTTP_REFERER');
            $pathTrust = null;
            $currentPathRequest = $request->getOriginalPathInfo();
            $currentPathRequest = ltrim($currentPathRequest, '/');

            $urlRewriteConnection = $this->urlRewriteResource->getConnection();
            $selectq = $urlRewriteConnection->select()->from(
                $urlRewriteConnection->getTableName('url_rewrite'),
                ['url_rewrite_id', 'entity_type', 'entity_id', 'target_path']
            )->where('request_path = ?', $currentPathRequest)->where('store_id = ?', $storeConditionId)->limit(1);
            $urlRewriteData = $urlRewriteConnection->fetchAll($selectq);
            if ($urlRewriteData && isset($urlRewriteData[0]) && is_array($urlRewriteData[0])) {
                $entityType = $urlRewriteData[0]['entity_type'];
                $entityId = $urlRewriteData[0]['entity_id'];
                $target = $urlRewriteData[0]['target_path'];

                $selecty = $urlRewriteConnection->select()->from(
                    $urlRewriteConnection->getTableName('url_rewrite'),
                    ['request_path']
                )->where(
                    'entity_type = ?', $entityType
                )->where(
                    'store_id = ?', $toStore->getId()
                )->where(
                    'target_path = ?', $target
                )->where('entity_id = ?', $entityId)->limit(1);
                $urlRewriteNextStore = $urlRewriteConnection->fetchAll($selecty);
                if ($urlRewriteNextStore && isset($urlRewriteNextStore[0]) && is_array($urlRewriteNextStore[0])) {
                    $pathTrust = $urlRewriteNextStore[0]['request_path'];
                }
                if ($pathTrust) {
                    $toUrl = trim($toStore->getBaseUrl(), '/') . '/' . $pathTrust;
                    return $toUrl;
                }
            } else {
                return $toStore->getBaseUrl();
            }
        }
        if ($toStore->getCode() === $defaultStore->getCode()) {
            $toUrl = $toStore->getBaseUrl();
            if ($currentPath && strlen($currentPath) && $currentPath !== '/') {
                $toUrl = trim($toStore->getBaseUrl(), '/') . '/' . $currentPath;
            }
            return $this->storeSwitcher->switch(
                $fromStore,
                $toStore,
                $toUrl
            );
        }
        $this->urlBuilder->setScope($toStore->getId());
        $targetUrl = $this->urlBuilder->getUrl(
            $currentPath,
            [
                '_current' => false,
                '_nosid' => true,
                '_query' => [
                    StoreManagerInterface::PARAM_NAME => $toStore
                ]
            ]
        );
        $href = $this->urlBuilder->getUrl(
            'stores/store/redirect',
            [
                '_current' => false,
                '_nosid' => true,
                '_query' => $this->prepareRequestQuery($toStore->getCode(), $targetUrl)
            ]
        );
        return $href;
    }

    protected function getPathByProduct($entityId, $storeId, $hasHtmlExtends)
    {
        $product = null;
        try {
            $product = $this->productRepository->getById($entityId, false, $storeId);
        } catch (\Exception $exception) {
            // Do nothing
        }
        if ($product) {
            $currentPath = $product->getUrlKey();
            $currentPath = $hasHtmlExtends ? $currentPath . '.html' : $currentPath;
            return $currentPath;
        }
        return false;
    }

    protected function getPathByCategory($entityId, $storeId, $hasHtmlExtends)
    {
        $category = null;
        try {
            $category = $this->categoryRepository->get($entityId, $storeId);
        } catch (\Exception $exception) {
            // Do nothing
        }
        if ($category) {
            $currentPath = $category->getUrlKey();
            $currentPath = $hasHtmlExtends ? $currentPath . '.html' : $currentPath;
            return $currentPath;
        }
        return false;
    }

    /**
     * Prepare request query
     * param $store is store code
     *
     * @param string $store
     * @param string $href
     * @return array
     */
    private function prepareRequestQuery($store, $href)
    {
        $storeView = $this->storeManager->getDefaultStoreView();
        $query = [
            StoreManagerInterface::PARAM_NAME => $store,
            \Magento\Framework\App\ActionInterface::PARAM_NAME_URL_ENCODED => $this->encoder->encode($href)
        ];
        if (null !== $storeView && $storeView->getCode() !== $store) {
            $query['___from_store'] = $storeView->getCode();
        }

        return $query;
    }
}
